"""Factory classes for MongoDB models."""

import random

import factory
from faker import Faker

from mork.edx.mongo.models import Comment, CommentThread


class CommentBaseFactory(factory.mongoengine.MongoEngineFactory):
    """Base factory for MongoDB comment objects."""

    class Params:
        """Factory parameter to pick username from a list."""

        usernames = []

    votes = {}
    visible = factory.Faker("pybool")
    abuse_flaggers = []
    historical_abuse_flaggers = []
    thread_type = factory.Faker("pystr")
    context = factory.Faker("pystr")
    comment_count = factory.Faker("pyint")
    at_position_list = []
    body = factory.Faker("text")
    course_id = factory.Faker("pystr")
    commentable_id = factory.Faker("pystr")
    anonymous = factory.Faker("pybool")
    anonymous_to_peers = factory.Faker("pybool")
    closed = factory.Faker("pybool")
    author_id = factory.Faker("pyint")
    author_username = factory.LazyAttribute(
        lambda f: (
            random.choice(f.usernames) if f.usernames else Faker().pystr()  # noqa: S311
        )
    )
    updated_at = factory.Faker("date_time")
    created_at = factory.Faker("date_time")
    last_activity_at = factory.Faker("date_time")


class CommentFactory(CommentBaseFactory):
    """Factory for the `Comment` document type."""

    class Meta:
        """Factory configuration."""

        model = Comment

    type = "Comment"


class CommentThreadFactory(CommentBaseFactory):
    """Factory for the `CommentThread` document type."""

    class Meta:
        """Factory configuration."""

        model = CommentThread

    type = "CommentThread"
    title = factory.Faker("sentence")



"""Create users tables

Revision ID: b61049edae38
Revises: 976e462b45b6
Create Date: 2024-11-08 10:38:01.575241

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "b61049edae38"
down_revision: Union[str, None] = "976e462b45b6"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("username", sa.String(length=254), nullable=False),
        sa.Column("edx_user_id", sa.Integer(), nullable=False),
        sa.Column("email", sa.String(length=254), nullable=False),
        sa.Column(
            "reason",
            sa.Enum("USER_REQUESTED", "GDPR", name="deletion_reason"),
            nullable=False,
        ),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("edx_user_id"),
        sa.UniqueConstraint("username"),
    )
    op.create_table(
        "user_service_statuses",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("user_id", sa.UUID(), nullable=False),
        sa.Column(
            "service_name",
            sa.Enum("ASHLEY", "EDX", "BREVO", "JOANIE", name="service_name"),
            nullable=False,
        ),
        sa.Column(
            "status",
            sa.Enum("TO_DELETE", "DELETING", "DELETED", name="deletion_status"),
            nullable=False,
        ),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["user_id"], ["users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "service_name", name="uq_record_service"),
    )
    op.create_index(
        "idx_service_status",
        "user_service_statuses",
        ["service_name", "status"],
        unique=False,
    )
    op.create_index(
        "idx_user_service_status",
        "user_service_statuses",
        ["user_id", "service_name", "status"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("idx_user_service_status", table_name="user_service_statuses")
    op.drop_index("idx_service_status", table_name="user_service_statuses")
    op.drop_table("user_service_statuses")
    op.drop_table("users")
    # ### end Alembic commands ###



"""Mork edx verify models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class VerifyStudentHistoricalverificationdeadline(Base):
    """Model for the `verify_student_historicalverificationdeadline` table."""

    __tablename__ = "verify_student_historicalverificationdeadline"
    __table_args__ = (
        ForeignKeyConstraint(
            ["history_user_id"],
            ["auth_user.id"],
        ),
    )
    id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    modified: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    course_key: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    deadline: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    history_id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    history_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    history_user_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    history_type: Mapped[str] = mapped_column(String(1), nullable=False)
    deadline_is_explicit: Mapped[int] = mapped_column(INTEGER(1), nullable=False)

    history_user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="verify_student_historicalverificationdeadline"
    )


class VerifyStudentSoftwaresecurephotoverification(Base):
    """Model for the `verify_student_softwaresecurephotoverification` table."""

    __tablename__ = "verify_student_softwaresecurephotoverification"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        ForeignKeyConstraint(
            ["reviewing_user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    status: Mapped[str] = mapped_column(String(100), nullable=False)
    status_changed: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    face_image_url: Mapped[str] = mapped_column(String(255), nullable=False)
    photo_id_image_url: Mapped[str] = mapped_column(String(255), nullable=False)
    receipt_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    created_at: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    updated_at: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    submitted_at: Mapped[datetime.datetime] = mapped_column(DateTime, index=True)
    reviewing_user_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    reviewing_service: Mapped[str] = mapped_column(String(255), nullable=False)
    error_msg: Mapped[str] = mapped_column(TEXT, nullable=False)
    error_code: Mapped[str] = mapped_column(String(50), nullable=False)
    photo_id_key: Mapped[str] = mapped_column(TEXT, nullable=False)
    display: Mapped[int] = mapped_column(INTEGER(1), nullable=False, index=True)
    copy_id_photo_from_id: Mapped[int] = mapped_column(INTEGER(11), index=True)

    reviewing_user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser",
        primaryjoin="VerifyStudentSoftwaresecurephotoverification.reviewing_user_id == AuthUser.id",  # noqa: E501
        back_populates="verify_student_softwaresecurephotoverification_reviewing_user",
    )
    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser",
        primaryjoin="VerifyStudentSoftwaresecurephotoverification.user_id == AuthUser.id",  # noqa: E501
        back_populates="verify_student_softwaresecurephotoverification_user",
    )



"""Tests of the MongoDB related CRUD functions."""

from mork.conf import settings
from mork.edx.mongo import crud
from mork.edx.mongo.factories import CommentFactory, CommentThreadFactory
from mork.edx.mongo.models import Comment, CommentThread


def test_edx_mongo_crud_anonymize_comments(edx_mongo_db):
    """Test the `anonymize_comments` method."""
    username = "JohnDoe"
    CommentFactory.create_batch(3, author_username=username)
    CommentThreadFactory.create_batch(4, author_username=username)

    assert Comment.objects(author_username=username).count() == 3
    assert CommentThread.objects(author_username=username).count() == 4

    count = crud.anonymize_comments(username)

    assert count == 7
    comment = Comment.objects().first()
    assert comment.author_username == "[deleted]"
    assert comment.body == "[deleted]"
    assert comment.author_id == settings.EDX_FORUM_PLACEHOLDER_USER_ID
    assert comment.anonymous

    comment_thread = CommentThread.objects().first()
    assert comment_thread.author_username == "[deleted]"
    assert comment_thread.title == "[deleted]"
    assert comment_thread.body == "[deleted]"
    assert comment_thread.author_id == settings.EDX_FORUM_PLACEHOLDER_USER_ID
    assert comment_thread.anonymous



"""Mork users models."""

import enum
from typing import List
from uuid import uuid4

from sqlalchemy import (
    Enum,
    ForeignKey,
    Index,
    Integer,
    String,
    UniqueConstraint,
)
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import Mapped, mapped_column, relationship

from mork.models import Base


class ServiceName(str, enum.Enum):
    """Enum for service names."""

    ASHLEY = "ashley"
    EDX = "edx"
    BREVO = "brevo"
    JOANIE = "joanie"


class DeletionReason(str, enum.Enum):
    """Enum for the reason of the deletion request."""

    USER_REQUESTED = "user_requested"
    GDPR = "gdpr"


class DeletionStatus(str, enum.Enum):
    """Enum for deletion statuses."""

    TO_DELETE = "to_delete"
    DELETING = "deleting"
    DELETED = "deleted"
    PROTECTED = "protected"


class UserServiceStatus(Base):
    """Table for storing the user status for a service."""

    __tablename__ = "user_service_statuses"
    __table_args__ = (
        UniqueConstraint("user_id", "service_name", name="uq_record_service"),
        Index(
            "idx_user_service_status",
            "user_id",
            "service_name",
            "status",
        ),
        Index("idx_service_status", "service_name", "status"),
    )

    id: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, default=uuid4
    )
    user_id: Mapped[UUID] = mapped_column(
        ForeignKey("users.id", ondelete="CASCADE"), nullable=False
    )
    service_name: Mapped[ServiceName] = mapped_column(
        Enum(ServiceName, name="service_name"), nullable=False
    )
    status: Mapped[DeletionStatus] = mapped_column(
        Enum(DeletionStatus, name="deletion_status"), nullable=False, default=False
    )

    user: Mapped["User"] = relationship("User", back_populates="service_statuses")


class User(Base):
    """Table for storing the users to delete."""

    __tablename__ = "users"

    filtered_attrs = []

    id: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, default=uuid4
    )
    username: Mapped[str] = mapped_column(String(254), unique=True, nullable=False)
    edx_user_id: Mapped[int] = mapped_column(Integer(), unique=True)
    email: Mapped[str] = mapped_column(String(254), nullable=False)
    reason: Mapped[DeletionReason] = mapped_column(
        Enum(DeletionReason, name="deletion_reason"), nullable=False
    )

    service_statuses: Mapped[List[UserServiceStatus]] = relationship(
        "UserServiceStatus",
        cascade="all, delete-orphan",
        back_populates="user",
    )



"""Tests of the CRUD functions."""

from datetime import datetime, timedelta

import pytest
from faker import Faker

from mork.edx.mysql import crud
from mork.edx.mysql.factories.auth import EdxAuthtokenTokenFactory, EdxAuthUserFactory
from mork.edx.mysql.factories.certificates import (
    EdxCertificatesCertificatehtmlviewconfigurationFactory,
)
from mork.edx.mysql.factories.contentstore import (
    EdxContentstoreVideouploadconfigFactory,
)
from mork.edx.mysql.factories.course import (
    EdxCourseActionStateCoursererunstateFactory,
    EdxCourseCreatorsCoursecreatorFactory,
)
from mork.edx.mysql.factories.dark import EdxDarkLangDarklangconfigFactory
from mork.edx.mysql.factories.student import (
    EdxStudentCourseenrollmentallowedFactory,
)
from mork.edx.mysql.factories.util import EdxUtilRatelimitconfigurationFactory
from mork.edx.mysql.factories.verify import (
    EdxVerifyStudentHistoricalverificationdeadlineFactory,
)
from mork.edx.mysql.models.auth import AuthUser
from mork.edx.mysql.models.base import Base
from mork.exceptions import UserNotFound, UserProtected


def test_edx_crud_get_inactive_users_count(edx_mysql_db):
    """Test the `get_inactive_users_count` method."""
    # 3 users that did not log in for 3 years
    EdxAuthUserFactory.create_batch(
        3, last_login=Faker().date_time_between(end_date="-3y")
    )
    # 4 users that logged in recently
    EdxAuthUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get count of users inactive for more than 3 years
    users_count = crud.get_inactive_users_count(edx_mysql_db.session, threshold_date)

    assert users_count == 3


def test_edx_crud_get_inactive_users_count_empty(edx_mysql_db):
    """Test the `get_inactive_users_count` method with no inactive users."""
    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get count of users inactive for more than 3 years
    users_count = crud.get_inactive_users_count(edx_mysql_db.session, threshold_date)

    assert users_count == 0


def test_edx_crud_get_inactive_users(edx_mysql_db):
    """Test the `get_inactive_users` method."""

    # 3 users that did not log in for 3 years
    inactive_users = EdxAuthUserFactory.create_batch(
        3, last_login=Faker().date_time_between(end_date="-3y")
    )
    # 4 users that logged in recently
    EdxAuthUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get all users inactive for more than 3 years
    users = crud.get_inactive_users(
        edx_mysql_db.session, threshold_date, offset=0, limit=9
    )

    assert len(users) == 3
    assert users == inactive_users


def test_edx_crud_get_inactive_users_empty(edx_mysql_db):
    """Test the `get_inactive_users` method with no inactive users."""

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    users = crud.get_inactive_users(
        edx_mysql_db.session, threshold_date, offset=0, limit=9
    )

    assert users == []


def test_edx_crud_get_inactive_users_slice(edx_mysql_db):
    """Test the `get_inactive_users` method with a slice."""
    # 3 users that did not log in for 3 years
    inactive_users = EdxAuthUserFactory.create_batch(
        3, last_login=Faker().date_time_between(end_date="-3y")
    )
    # 4 users that logged in recently
    EdxAuthUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    # Get two users inactive for more than 3 years
    users = crud.get_inactive_users(
        edx_mysql_db.session, threshold_date, offset=0, limit=2
    )

    assert len(users) == 2
    assert users == inactive_users[:2]


def test_edx_crud_get_inactive_users_slice_empty(edx_mysql_db):
    """Test the `get_inactive_users` method with an empty slice ."""
    # 3 users that did not log in for 3 years
    EdxAuthUserFactory.create_batch(
        3, last_login=Faker().date_time_between(end_date="-3y")
    )
    # 4 users that logged in recently
    EdxAuthUserFactory.create_batch(
        4, last_login=Faker().date_time_between(start_date="-3y")
    )

    threshold_date = datetime.now() - timedelta(days=365 * 3)

    users = crud.get_inactive_users(
        edx_mysql_db.session, threshold_date, offset=4, limit=9
    )

    assert users == []


def test_edx_crud_get_user_missing(edx_mysql_db):
    """Test the `get_user` method with missing user in the database."""

    user = crud.get_user(session=edx_mysql_db.session, email="john_doe@example.com")
    assert user is None


def test_edx_crud_get_user(edx_mysql_db):
    """Test the `get_user` method."""
    email = "john_doe@example.com"

    EdxAuthUserFactory.create_batch(1, email=email)

    user = crud.get_user(session=edx_mysql_db.session, email=email)
    assert user.email == email


def test_edx_crud_delete_user_missing(edx_mysql_db):
    """Test the `delete_user` method with missing user in the database."""

    email = "john_doe@example.com"

    with pytest.raises(UserNotFound, match="User does not exist"):
        crud.delete_user(edx_mysql_db.session, email=email)


def test_edx_crud_delete_user(edx_mysql_db):
    """Test the `delete_user` method."""
    email = "john_doe@example.com"
    EdxAuthUserFactory.create_batch(1, email=email)
    EdxStudentCourseenrollmentallowedFactory.create_batch(3, email=email)

    # Get all related tables that have foreign key constraints on the parent table
    related_tables = [
        "auth_userprofile",
        "auth_user_groups",
        "auth_registration",
        "bulk_email_courseemail",
        "bulk_email_optout",
        "certificates_generatedcertificate",
        "course_groups_courseusergroup_users",
        "course_groups_cohortmembership",
        "courseware_offlinecomputedgrade",
        "courseware_studentmodule",
        "courseware_studentmodulehistory",
        "courseware_xmodulestudentinfofield",
        "courseware_xmodulestudentprefsfield",
        "django_comment_client_role_users",
        "instructor_task_instructortask",
        "notify_settings",
        "payment_useracceptance",
        "proctoru_proctoruexam",
        "proctoru_proctoruuser",
        "student_anonymoususerid",
        "student_courseaccessrole",
        "student_courseenrollment",
        "student_courseenrollmentallowed",
        "student_courseenrollmentattribute",
        "student_historicalcourseenrollment",
        "student_languageproficiency",
        "student_loginfailures",
        "student_manualenrollmentaudit",
        "student_pendingemailchange",
        "student_userstanding",
        "user_api_userpreference",
        "verify_student_softwaresecurephotoverification",
    ]

    for table_name in related_tables:
        table = Base.metadata.tables[table_name]
        assert edx_mysql_db.session.query(table).count() > 0

    crud.delete_user(edx_mysql_db.session, email="john_doe@example.com")

    # Ensure the parent table is empty
    assert edx_mysql_db.session.query(AuthUser).count() == 0

    # Ensure the deletion has cascaded properly on children table
    for table_name in related_tables:
        table = Base.metadata.tables[table_name]
        assert edx_mysql_db.session.query(table).count() == 0


def test_edx_crud_delete_user_protected_table(edx_mysql_db):
    """Test the `delete_user` method on a user with entries in a protected tables."""
    email = "john_doe@example.com"
    EdxAuthUserFactory.create_batch(
        1,
        email=email,
        with_protected_tables=True,
    )

    # Get all related tables that have foreign key constraints on the parent table
    protected_tables = [
        "authtoken_token",
        "certificates_certificatehtmlviewconfiguration",
        "contentstore_videouploadconfig",
        "course_action_state_coursererunstate",
        "course_creators_coursecreator",
        "dark_lang_darklangconfig",
        "util_ratelimitconfiguration",
        "verify_student_historicalverificationdeadline",
        "wiki_article",
        "wiki_articlerevision",
    ]

    for table_name in protected_tables:
        table = Base.metadata.tables[table_name]
        assert edx_mysql_db.session.query(table).count() > 0

    with pytest.raises(
        UserProtected,
        match="User is linked to a protected table and cannot be deleted",
    ):
        crud.delete_user(edx_mysql_db.session, email=email)

    # Ensure the parent table is empty
    assert edx_mysql_db.session.query(AuthUser).count() > 0

    # Ensure the deletion has not cascaded on any protected children table
    for table_name in protected_tables:
        table = Base.metadata.tables[table_name]
        assert edx_mysql_db.session.query(table).count() > 0


def test_edx_crud_has_protected_children(edx_mysql_db):
    """Test the `_has_protected_children` method."""
    id = 123
    EdxAuthtokenTokenFactory.create(user_id=id)
    EdxCertificatesCertificatehtmlviewconfigurationFactory.create(changed_by_id=id)
    EdxContentstoreVideouploadconfigFactory.create(changed_by_id=id)
    EdxCourseActionStateCoursererunstateFactory.create(created_user_id=id)
    EdxCourseActionStateCoursererunstateFactory.create(updated_user_id=id)
    EdxCourseCreatorsCoursecreatorFactory.create(user_id=id)
    EdxDarkLangDarklangconfigFactory.create(changed_by_id=id)
    EdxUtilRatelimitconfigurationFactory.create(changed_by_id=id)
    EdxVerifyStudentHistoricalverificationdeadlineFactory.create(history_user_id=id)

    user_id = 456
    assert not crud._has_protected_children(edx_mysql_db.session, user_id)

    EdxCourseActionStateCoursererunstateFactory.create(updated_user_id=user_id)

    assert crud._has_protected_children(edx_mysql_db.session, user_id)



"""Fixtures for the mork app."""

from typing import AsyncIterator

import pytest
from httpx import ASGITransport, AsyncClient

from mork.api import app
from mork.conf import settings

transport = ASGITransport(app=app)


@pytest.fixture
async def http_client() -> AsyncIterator[AsyncClient]:
    """Handle application lifespan while yielding asynchronous HTTP client."""
    async with AsyncClient(transport=transport, base_url=settings.SERVER_URL) as client:
        yield client



"""Mork edx instructor models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class InstructorTaskInstructortask(Base):
    """Model for the `instructor_task_instructortask` table."""

    __tablename__ = "instructor_task_instructortask"
    __table_args__ = (
        ForeignKeyConstraint(
            ["requester_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    task_type: Mapped[str] = mapped_column(String(50), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    task_key: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    task_input: Mapped[str] = mapped_column(String(255), nullable=False)
    task_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    task_state: Mapped[str] = mapped_column(String(50), index=True)
    task_output: Mapped[str] = mapped_column(String(1024))
    requester_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(DateTime)
    updated: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    subtasks: Mapped[str] = mapped_column(TEXT, nullable=False)

    requester: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="instructor_task_instructortask"
    )



"""Factory classes for users."""

import factory

from mork.models.users import (
    DeletionReason,
    DeletionStatus,
    ServiceName,
    User,
    UserServiceStatus,
)


class UserServiceStatusFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory for generating UserServiceStatus instances."""

    class Meta:
        """Factory configuration."""

        model = UserServiceStatus

    user_id = factory.Sequence(lambda n: n + 1)
    service_name = ServiceName
    status = DeletionStatus.TO_DELETE


class UserFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Factory for generating fake users."""

    class Meta:
        """Factory configuration."""

        model = User

    username = factory.Faker("pystr", max_chars=15)
    edx_user_id = factory.Sequence(lambda n: n)
    email = factory.Faker("email")
    reason = DeletionReason.GDPR

    @factory.post_generation
    def service_statuses(
        self,
        create: bool,
        extracted: dict[ServiceName, DeletionStatus] | None,
        **kwargs,
    ):
        """Post-generation hook to create UserServiceStatus for all services."""
        if not create:
            return

        service_states = {service: DeletionStatus.TO_DELETE for service in ServiceName}

        if isinstance(extracted, dict):
            service_states.update(extracted)

        for service, status in service_states.items():
            UserServiceStatusFactory(
                user=self,
                service_name=service,
                status=status,
                **kwargs,
            )



"""Exceptions for Mork."""


class EmailSendError(Exception):
    """Raised when an error occurs when sending an email."""


class UserDeleteError(Exception):
    """Raised when an error occurs when deleting a user."""


class UserNotFound(Exception):
    """Raised when a user has not been found."""


class UserStatusError(Exception):
    """Raised when an error occurs when checking a user status."""


class UserProtected(Exception):
    """Raised when a user is associated with an entry in a protected table."""



"""Add protected to enum type

Revision ID: f2802e96b9d9
Revises: b61049edae38
Create Date: 2024-11-27 17:40:32.627819

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from alembic_postgresql_enum import TableReference

# revision identifiers, used by Alembic.
revision: str = "f2802e96b9d9"
down_revision: Union[str, None] = "b61049edae38"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.sync_enum_values(
        enum_schema="public",
        enum_name="deletion_status",
        new_values=["TO_DELETE", "DELETING", "DELETED", "PROTECTED"],
        affected_columns=[
            TableReference(
                table_schema="public",
                table_name="user_service_statuses",
                column_name="status",
            )
        ],
        enum_values_to_rename=[],
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.sync_enum_values(
        enum_schema="public",
        enum_name="deletion_status",
        new_values=["TO_DELETE", "DELETING", "DELETED"],
        affected_columns=[
            TableReference(
                table_schema="public",
                table_name="user_service_statuses",
                column_name="status",
            )
        ],
        enum_values_to_rename=[],
    )
    # ### end Alembic commands ###



"""Mork edx auth models."""

import datetime
from typing import List, Optional

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base
from .bulk import BulkEmailCourseemail, BulkEmailOptout
from .certificates import (
    CertificatesCertificatehtmlviewconfiguration,
    CertificatesGeneratedcertificate,
)
from .contentstore import ContentstoreVideouploadconfig
from .course import (
    CourseActionStateCoursererunstate,
    CourseCreatorsCoursecreator,
    CourseGroupsCohortmembership,
    CourseGroupsCourseusergroupUsers,
)
from .courseware import (
    CoursewareOfflinecomputedgrade,
    CoursewareStudentmodule,
    CoursewareXmodulestudentinfofield,
    CoursewareXmodulestudentprefsfield,
)
from .dark import DarkLangDarklangconfig
from .django import DjangoCommentClientRoleUsers
from .instructor import InstructorTaskInstructortask
from .notify import NotifySetting
from .payment import PaymentUseracceptance
from .proctoru import ProctoruProctoruexam, ProctoruProctoruuser
from .student import (
    StudentAnonymoususerid,
    StudentCourseaccessrole,
    StudentCourseenrollment,
    StudentHistoricalcourseenrollment,
    StudentLanguageproficiency,
    StudentLoginfailure,
    StudentPendingemailchange,
    StudentUserstanding,
)
from .user import UserApiUserpreference
from .util import UtilRatelimitconfiguration
from .verify import (
    VerifyStudentHistoricalverificationdeadline,
    VerifyStudentSoftwaresecurephotoverification,
)
from .wiki import WikiArticle, WikiArticlerevision


class AuthUser(Base):
    """Model for the `auth_user` table."""

    __tablename__ = "auth_user"
    __table_args__ = (
        Index("email", "email", unique=True),
        Index("username", "username", unique=True),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    username: Mapped[str] = mapped_column(String(30))
    first_name: Mapped[str] = mapped_column(String(30))
    last_name: Mapped[str] = mapped_column(String(30))
    email: Mapped[str] = mapped_column(String(254))
    password: Mapped[str] = mapped_column(String(128))
    is_staff: Mapped[int] = mapped_column(INTEGER(1))
    is_active: Mapped[int] = mapped_column(INTEGER(1))
    is_superuser: Mapped[int] = mapped_column(INTEGER(1))
    date_joined: Mapped[datetime.datetime] = mapped_column(DateTime)
    last_login: Mapped[Optional[datetime.datetime]] = mapped_column(DateTime)

    auth_registration: Mapped["AuthRegistration"] = relationship(
        "AuthRegistration",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    authtoken_token: Mapped["AuthtokenToken"] = relationship(
        "AuthtokenToken",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    auth_userprofile: Mapped["AuthUserprofile"] = relationship(
        "AuthUserprofile",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    auth_user_groups: Mapped["AuthUserGroups"] = relationship(
        "AuthUserGroups",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    bulk_email_courseemail: Mapped[List["BulkEmailCourseemail"]] = relationship(
        "BulkEmailCourseemail",
        back_populates="sender",
        cascade="all, delete-orphan",
    )
    bulk_email_optout: Mapped[List["BulkEmailOptout"]] = relationship(
        "BulkEmailOptout",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    certificates_certificatehtmlviewconfiguration: Mapped[
        List["CertificatesCertificatehtmlviewconfiguration"]
    ] = relationship(
        "CertificatesCertificatehtmlviewconfiguration",
        back_populates="changed_by",
        cascade="all, delete-orphan",
    )
    certificates_generatedcertificate: Mapped[
        List["CertificatesGeneratedcertificate"]
    ] = relationship(
        "CertificatesGeneratedcertificate",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    contentstore_videouploadconfig: Mapped[List["ContentstoreVideouploadconfig"]] = (
        relationship(
            "ContentstoreVideouploadconfig",
            back_populates="changed_by",
            cascade="all, delete-orphan",
        )
    )
    course_action_state_coursererunstate_created_user: Mapped[
        List["CourseActionStateCoursererunstate"]
    ] = relationship(
        "CourseActionStateCoursererunstate",
        back_populates="created_user",
        cascade="all, delete-orphan",
        foreign_keys=[CourseActionStateCoursererunstate.created_user_id],
    )
    course_action_state_coursererunstate_updated_user: Mapped[
        List["CourseActionStateCoursererunstate"]
    ] = relationship(
        "CourseActionStateCoursererunstate",
        back_populates="updated_user",
        cascade="all, delete-orphan",
        foreign_keys=[CourseActionStateCoursererunstate.updated_user_id],
    )
    course_groups_courseusergroup_users: Mapped[
        List["CourseGroupsCourseusergroupUsers"]
    ] = relationship(
        "CourseGroupsCourseusergroupUsers",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    course_groups_cohortmembership: Mapped[List["CourseGroupsCohortmembership"]] = (
        relationship(
            "CourseGroupsCohortmembership",
            back_populates="user",
            cascade="all, delete-orphan",
        )
    )
    course_creators_coursecreator: Mapped["CourseCreatorsCoursecreator"] = relationship(
        "CourseCreatorsCoursecreator",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    courseware_offlinecomputedgrade: Mapped[List["CoursewareOfflinecomputedgrade"]] = (
        relationship(
            "CoursewareOfflinecomputedgrade",
            back_populates="user",
            cascade="all, delete-orphan",
        )
    )
    courseware_studentmodule: Mapped[List["CoursewareStudentmodule"]] = relationship(
        "CoursewareStudentmodule",
        back_populates="student",
        cascade="all, delete-orphan",
    )
    courseware_xmodulestudentinfofield: Mapped[
        List["CoursewareXmodulestudentinfofield"]
    ] = relationship(
        "CoursewareXmodulestudentinfofield",
        back_populates="student",
        cascade="all, delete-orphan",
    )
    courseware_xmodulestudentprefsfield: Mapped[
        List["CoursewareXmodulestudentprefsfield"]
    ] = relationship(
        "CoursewareXmodulestudentprefsfield",
        back_populates="student",
        cascade="all, delete-orphan",
    )
    django_comment_client_role_users: Mapped[List["DjangoCommentClientRoleUsers"]] = (
        relationship(
            "DjangoCommentClientRoleUsers",
            back_populates="user",
            cascade="all, delete-orphan",
        )
    )
    dark_lang_darklangconfig: Mapped[List["DarkLangDarklangconfig"]] = relationship(
        "DarkLangDarklangconfig",
        back_populates="changed_by",
        cascade="all, delete-orphan",
    )
    instructor_task_instructortask: Mapped[List["InstructorTaskInstructortask"]] = (
        relationship(
            "InstructorTaskInstructortask",
            back_populates="requester",
            cascade="all, delete-orphan",
        )
    )
    notify_settings: Mapped[List["NotifySetting"]] = relationship(
        "NotifySetting",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    payment_useracceptance: Mapped[List["PaymentUseracceptance"]] = relationship(
        "PaymentUseracceptance",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    proctoru_proctoruexam: Mapped[List["ProctoruProctoruexam"]] = relationship(
        "ProctoruProctoruexam",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    proctoru_proctoruuser: Mapped["ProctoruProctoruuser"] = relationship(
        "ProctoruProctoruuser",
        back_populates="student",
        cascade="all, delete-orphan",
    )
    student_anonymoususerid: Mapped[List["StudentAnonymoususerid"]] = relationship(
        "StudentAnonymoususerid",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    student_courseaccessrole: Mapped[List["StudentCourseaccessrole"]] = relationship(
        "StudentCourseaccessrole",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    student_courseenrollment: Mapped[List["StudentCourseenrollment"]] = relationship(
        "StudentCourseenrollment",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    student_historicalcourseenrollment_history_user: Mapped[
        List["StudentHistoricalcourseenrollment"]
    ] = relationship(
        "StudentHistoricalcourseenrollment",
        foreign_keys=[StudentHistoricalcourseenrollment.history_user_id],
        back_populates="history_user",
        cascade="all, delete-orphan",
    )
    student_historicalcourseenrollment_user: Mapped[
        List["StudentHistoricalcourseenrollment"]
    ] = relationship(
        "StudentHistoricalcourseenrollment",
        foreign_keys=[StudentHistoricalcourseenrollment.user_id],
        back_populates="user",
        cascade="all, delete-orphan",
    )
    student_loginfailures: Mapped[List["StudentLoginfailure"]] = relationship(
        "StudentLoginfailure",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    student_pendingemailchange: Mapped["StudentPendingemailchange"] = relationship(
        "StudentPendingemailchange",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    student_userstanding_changed_by: Mapped["StudentUserstanding"] = relationship(
        "StudentUserstanding",
        foreign_keys=[StudentUserstanding.changed_by_id],
        primaryjoin="StudentUserstanding.changed_by_id == AuthUser.id",
        back_populates="changed_by",
        cascade="all, delete-orphan",
    )
    student_userstanding_user: Mapped["StudentUserstanding"] = relationship(
        "StudentUserstanding",
        foreign_keys=[StudentUserstanding.user_id, StudentUserstanding.changed_by_id],
        primaryjoin="StudentUserstanding.user_id == AuthUser.id",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    user_api_userpreference: Mapped[List["UserApiUserpreference"]] = relationship(
        "UserApiUserpreference",
        back_populates="user",
        cascade="all, delete-orphan",
    )
    util_ratelimitconfiguration: Mapped[List["UtilRatelimitconfiguration"]] = (
        relationship(
            "UtilRatelimitconfiguration",
            back_populates="changed_by",
            cascade="all, delete-orphan",
        )
    )
    verify_student_historicalverificationdeadline: Mapped[
        List["VerifyStudentHistoricalverificationdeadline"]
    ] = relationship(
        "VerifyStudentHistoricalverificationdeadline",
        back_populates="history_user",
        cascade="all, delete-orphan",
    )
    verify_student_softwaresecurephotoverification_reviewing_user: Mapped[
        List["VerifyStudentSoftwaresecurephotoverification"]
    ] = relationship(
        "VerifyStudentSoftwaresecurephotoverification",
        foreign_keys=[VerifyStudentSoftwaresecurephotoverification.reviewing_user_id],
        primaryjoin="VerifyStudentSoftwaresecurephotoverification.reviewing_user_id == AuthUser.id",  # noqa: E501
        back_populates="reviewing_user",
        cascade="all, delete-orphan",
    )
    verify_student_softwaresecurephotoverification_user: Mapped[
        List["VerifyStudentSoftwaresecurephotoverification"]
    ] = relationship(
        "VerifyStudentSoftwaresecurephotoverification",
        foreign_keys=[VerifyStudentSoftwaresecurephotoverification.user_id],
        primaryjoin="VerifyStudentSoftwaresecurephotoverification.user_id == AuthUser.id",  # noqa: E501
        back_populates="user",
        cascade="all, delete-orphan",
    )
    wiki_article: Mapped[List["WikiArticle"]] = relationship(
        "WikiArticle",
        back_populates="owner",
        cascade="all, delete-orphan",
    )
    wiki_articlerevision: Mapped[List["WikiArticlerevision"]] = relationship(
        "WikiArticlerevision",
        back_populates="user",
        cascade="all, delete-orphan",
    )


class AuthUserGroups(Base):
    """Model for the `auth_user_groups`table."""

    __tablename__ = "auth_user_groups"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index("user_id_auth_user_groups", "user_id", "group_id", unique=True),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped["AuthUser"] = mapped_column(INTEGER(11), nullable=False, index=True)
    group_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)

    user: Mapped["AuthUser"] = relationship(
        "AuthUser", back_populates="auth_user_groups"
    )


class AuthRegistration(Base):
    """Model for the `auth_registration` table."""

    __tablename__ = "auth_registration"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )
    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11))
    activation_key: Mapped[int] = mapped_column(String(32), nullable=False, unique=True)

    user: Mapped["AuthUser"] = relationship(
        "AuthUser", back_populates="auth_registration"
    )


class AuthtokenToken(Base):
    """Model for the `authtoken_token` table."""

    __tablename__ = "authtoken_token"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )
    key: Mapped[str] = mapped_column(String(40), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11))
    created: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)

    user: Mapped["AuthUser"] = relationship(
        "AuthUser", back_populates="authtoken_token"
    )


class AuthUserprofile(Base):
    """Model for the `auth_userprofile` table."""

    __tablename__ = "auth_userprofile"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
            name="user_id_refs_id_3daaa960628b4c11",
        ),
        Index("auth_userprofile_52094d6e", "name"),
        Index("auth_userprofile_551e365c", "level_of_education"),
        Index("auth_userprofile_8a7ac9ab", "language"),
        Index("auth_userprofile_b54954de", "location"),
        Index("auth_userprofile_d85587", "year_of_birth"),
        Index("auth_userprofile_fca3d292", "gender"),
        Index("user_id", "user_id", unique=True),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11))
    name: Mapped[str] = mapped_column(String(255))
    language: Mapped[str] = mapped_column(String(255))
    location: Mapped[str] = mapped_column(String(255))
    meta: Mapped[str] = mapped_column(TEXT)
    courseware: Mapped[str] = mapped_column(String(255))
    allow_certificate: Mapped[int] = mapped_column(INTEGER(1))
    gender: Mapped[Optional[str]] = mapped_column(String(6))
    mailing_address: Mapped[Optional[str]] = mapped_column(TEXT)
    year_of_birth: Mapped[Optional[int]] = mapped_column(INTEGER(11))
    level_of_education: Mapped[Optional[str]] = mapped_column(String(6))
    goals: Mapped[Optional[str]] = mapped_column(TEXT)
    country: Mapped[Optional[str]] = mapped_column(String(2))
    city: Mapped[Optional[str]] = mapped_column(TEXT)
    bio: Mapped[Optional[str]] = mapped_column(String(3000))
    profile_image_uploaded_at: Mapped[Optional[datetime.datetime]] = mapped_column(
        DateTime
    )

    user: Mapped["AuthUser"] = relationship(
        "AuthUser", back_populates="auth_userprofile"
    )

    student_languageproficiency: Mapped[List["StudentLanguageproficiency"]] = (
        relationship(
            "StudentLanguageproficiency",
            back_populates="user_profile",
            cascade="all, delete-orphan",
        )
    )



"""Tests for the Mork API '/tasks/' endpoints."""

from unittest.mock import Mock, patch

import pytest
from httpx import AsyncClient


@pytest.mark.anyio
async def test_tasks_auth(http_client: AsyncClient):
    """Test required authentication for tasks endpoints."""
    # FastAPI returns a 403 error (instead of a 401 error) if no API token is given
    # see https://github.com/tiangolo/fastapi/discussions/9130
    assert (await http_client.post("/v1/tasks/")).status_code == 403
    assert (await http_client.options("/v1/tasks/")).status_code == 403
    assert (await http_client.get("/v1/tasks/1234/status")).status_code == 403


@pytest.mark.anyio
@pytest.mark.parametrize(
    "body_params",
    [
        {"type": "email_inactive_users", "dry_run": False},
        {"type": "email_inactive_users", "limit": 100, "dry_run": False},
        {
            "type": "email_user",
            "email": "johndoe@example.com",
            "username": "JohnDoe",
            "dry_run": False,
        },
        {"type": "delete_inactive_users", "dry_run": False},
        {"type": "delete_inactive_users", "limit": 100, "dry_run": False},
        {"type": "delete_user", "email": "johndoe@example.com", "dry_run": False},
        {"type": "email_inactive_users", "dry_run": True},
        {
            "type": "email_user",
            "email": "johndoe@example.com",
            "username": "JohnDoe",
            "dry_run": True,
        },
        {"type": "delete_inactive_users", "dry_run": True},
        {"type": "delete_user", "email": "johndoe@example.com", "dry_run": True},
    ],
)
@pytest.mark.parametrize(
    "tasks_endpoint",
    ["/v1/tasks/", "/v1/tasks"],
)
async def test_create_task(
    http_client: AsyncClient, auth_headers: dict, body_params: str, tasks_endpoint: str
):
    """Test creating a task with valid data."""

    celery_task = Mock()
    celery_task.delay.return_value.task_id = "1234"

    with patch.dict(
        "mork.api.v1.tasks.TASK_TYPE_TO_FUNC", {body_params["type"]: celery_task}
    ):
        response = await http_client.post(
            tasks_endpoint, headers=auth_headers, json=body_params
        )
        response_data = response.json()

        assert response.status_code == 202
        assert response_data.get("id")
        assert response_data.get("status") == "PENDING"
        assert (
            response.headers["location"] == f"/tasks/{response_data.get("id")}/status"
        )

        expected_params = {
            key: value for key, value in body_params.items() if key != "type"
        }
        celery_task.delay.assert_called_with(**expected_params)


@pytest.mark.anyio
async def test_create_task_invalid_type(http_client: AsyncClient, auth_headers: dict):
    """Test creating a task with an invalid type."""

    # Without a type
    with patch.dict("mork.api.v1.tasks.TASK_TYPE_TO_FUNC"):
        response = await http_client.post("/v1/tasks/", headers=auth_headers)

        assert response.status_code == 422

    # With a wrong type
    mock_task_create = {"type": "wrong_type"}

    celery_task = Mock()
    celery_task.delay.return_value.task_id = "1234"

    with patch.dict("mork.api.v1.tasks.TASK_TYPE_TO_FUNC"):
        response = await http_client.post(
            "/v1/tasks/", headers=auth_headers, json=mock_task_create
        )

        assert response.status_code == 422


@pytest.mark.anyio
@pytest.mark.parametrize(
    "task_type, params",
    [
        ("delete_user", {}),
        ("email_user", {"email": "johndoe@example.com"}),
        ("email_user", {"username": "JohnDoe"}),
    ],
)
async def test_create_task_missing_param(
    http_client: AsyncClient, auth_headers: dict, task_type: str, params: dict
):
    """Test creating a task with a missing parameter."""

    mock_task_create = {"type": task_type, **params}

    celery_task = Mock()
    celery_task.delay.return_value.task_id = "1234"

    with patch.dict("mork.api.v1.tasks.TASK_TYPE_TO_FUNC", {task_type: celery_task}):
        response = await http_client.post(
            "/v1/tasks/", headers=auth_headers, json=mock_task_create
        )
        response_data = response.json()

        assert response.status_code == 422
        assert response_data["detail"][0]["msg"] == "Field required"


@pytest.mark.anyio
@pytest.mark.parametrize(
    "tasks_endpoint",
    ["/v1/tasks/", "/v1/tasks"],
)
async def test_get_available_tasks(
    http_client: AsyncClient, auth_headers: dict, tasks_endpoint: str
):
    """Test getting available tasks."""
    response = await http_client.options(tasks_endpoint, headers=auth_headers)
    response_data = response.json()
    assert response.status_code == 200
    assert response.headers["allow"] == "POST"
    assert sorted(response_data.get("task_types")) == [
        "delete_inactive_users",
        "delete_user",
        "email_inactive_users",
        "email_user",
    ]


@pytest.mark.anyio
async def test_get_task_status(http_client: AsyncClient, auth_headers: dict):
    """Test getting task status."""

    task_id = "1234"
    celery_result = Mock(task_id)
    celery_result(task_id).state = "SUCCESS"

    with patch("mork.api.v1.tasks.AsyncResult", celery_result):
        response = await http_client.get(
            f"/v1/tasks/{task_id}/status",
            headers=auth_headers,
        )
        response_data = response.json()
        assert response.status_code == 200
        assert response_data.get("status") == "SUCCESS"



"""Tests of the edx models."""

from mork.edx.mysql.factories.auth import EdxAuthUserFactory, EdxAuthUserprofileFactory
from mork.edx.mysql.factories.certificates import (
    EdxCertificatesGeneratedCertificateFactory,
)
from mork.edx.mysql.factories.student import EdxStudentCourseenrollmentFactory
from mork.edx.mysql.factories.user import EdxUserApiUserpreferenceFactory


def test_edx_models_auth_user_safe_dict(edx_mysql_db):
    """Test the `safe_dict` method for the AuthUser model."""
    edx_auth_user = EdxAuthUserFactory()

    assert edx_auth_user.safe_dict() == {
        "id": edx_auth_user.id,
        "username": edx_auth_user.username,
        "first_name": edx_auth_user.first_name,
        "last_name": edx_auth_user.last_name,
        "email": edx_auth_user.email,
        "password": edx_auth_user.password,
        "is_staff": edx_auth_user.is_staff,
        "is_superuser": edx_auth_user.is_superuser,
        "is_active": edx_auth_user.is_active,
        "date_joined": edx_auth_user.date_joined,
        "last_login": edx_auth_user.last_login,
    }


def test_edx_models_auth_user_profile_safe_dict(edx_mysql_db):
    """Test the `safe_dict` method for the AuthUserprofile model."""
    edx_auth_user_profile = EdxAuthUserprofileFactory()

    assert edx_auth_user_profile.safe_dict() == {
        "id": edx_auth_user_profile.id,
        "user_id": edx_auth_user_profile.user_id,
        "name": edx_auth_user_profile.name,
        "language": edx_auth_user_profile.language,
        "location": edx_auth_user_profile.location,
        "meta": edx_auth_user_profile.meta,
        "courseware": edx_auth_user_profile.courseware,
        "allow_certificate": edx_auth_user_profile.allow_certificate,
        "gender": edx_auth_user_profile.gender,
        "mailing_address": edx_auth_user_profile.mailing_address,
        "year_of_birth": edx_auth_user_profile.year_of_birth,
        "level_of_education": edx_auth_user_profile.level_of_education,
        "goals": edx_auth_user_profile.goals,
        "country": edx_auth_user_profile.country,
        "city": edx_auth_user_profile.city,
        "bio": edx_auth_user_profile.bio,
        "profile_image_uploaded_at": edx_auth_user_profile.profile_image_uploaded_at,
    }


def test_edx_models_user_api_user_preference_safe_dict(edx_mysql_db):
    """Test the `safe_dict` method for the UserApiUserPreference model."""
    edx_user_api_user_preference = EdxUserApiUserpreferenceFactory()

    assert edx_user_api_user_preference.safe_dict() == {
        "id": edx_user_api_user_preference.id,
        "user_id": edx_user_api_user_preference.user_id,
        "key": edx_user_api_user_preference.key,
        "value": edx_user_api_user_preference.value,
    }


def test_edx_models_student_course_enrollment_safe_dict(edx_mysql_db):
    """Test the `safe_dict` method for the StudentCourseEnrollment model."""
    edx_student_course_enrollment = EdxStudentCourseenrollmentFactory()

    assert edx_student_course_enrollment.safe_dict() == {
        "id": edx_student_course_enrollment.id,
        "user_id": edx_student_course_enrollment.user_id,
        "course_id": edx_student_course_enrollment.course_id,
        "is_active": edx_student_course_enrollment.is_active,
        "mode": edx_student_course_enrollment.mode,
        "created": edx_student_course_enrollment.created,
    }


def test_edx_models_certificates_generated_certificate_safe_dict(edx_mysql_db):
    """Test the `safe_dict` method for the StudentCourseEnrollment model."""
    edx_certificates_generated_certificate = (
        EdxCertificatesGeneratedCertificateFactory()
    )

    assert edx_certificates_generated_certificate.safe_dict() == {
        "id": edx_certificates_generated_certificate.id,
        "user_id": edx_certificates_generated_certificate.user_id,
        "download_url": edx_certificates_generated_certificate.download_url,
        "grade": edx_certificates_generated_certificate.grade,
        "course_id": edx_certificates_generated_certificate.course_id,
        "key": edx_certificates_generated_certificate.key,
        "distinction": edx_certificates_generated_certificate.distinction,
        "status": edx_certificates_generated_certificate.status,
        "verify_uuid": edx_certificates_generated_certificate.verify_uuid,
        "download_uuid": edx_certificates_generated_certificate.download_uuid,
        "name": edx_certificates_generated_certificate.name,
        "created_date": edx_certificates_generated_certificate.created_date,
        "modified_date": edx_certificates_generated_certificate.modified_date,
        "error_reason": edx_certificates_generated_certificate.error_reason,
        "mode": edx_certificates_generated_certificate.mode,
    }



"""Factory base configuration."""

import factory
from faker import Faker
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from mork.edx.mysql.models.base import Base

faker = Faker()
engine = create_engine("sqlite+pysqlite:///:memory:", echo=False, pool_pre_ping=True)
Base.metadata.create_all(engine)

Session = scoped_session(sessionmaker())


class BaseSQLAlchemyModelFactory(factory.alchemy.SQLAlchemyModelFactory):
    """Base factory class for SQLAlchemy models."""

    class Meta:
        """Factory configuration."""

        abstract = True
        sqlalchemy_session = Session
        sqlalchemy_session_persistence = None



"""Mork edx MongoDB database connection."""

from mongoengine import connect, disconnect

from mork.conf import settings


class OpenEdxMongoDB:
    """Class to connect to the Open edX MongoDB database."""

    connection = None

    def __init__(self, connection=None):
        """Instantiate the MongoDB connection."""
        if connection is not None:
            self.connection = connection
        else:
            self.connection = connect(
                host=settings.EDX_MONGO_DB_HOST,
                username=settings.EDX_MONGO_DB_USER,
                password=settings.EDX_MONGO_DB_PASSWORD,
                db=settings.EDX_MONGO_DB_NAME,
            )

    def disconnect(self):
        """Close the connection with MongoDB."""
        disconnect()



"""API routes related to users."""

import logging
from typing import Annotated
from uuid import UUID

from fastapi import APIRouter, Body, Depends, HTTPException, Path, Query, status
from sqlalchemy import select, update
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session

from mork.auth import authenticate_api_key
from mork.db import get_session
from mork.models.users import (
    ServiceName,
    User,
    UserServiceStatus,
)
from mork.schemas.users import (
    DeletionStatus,
    UserRead,
    UserStatusRead,
    UserStatusUpdate,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/users", dependencies=[Depends(authenticate_api_key)])


@router.get("")
@router.get("/")
async def read_users(
    session: Annotated[Session, Depends(get_session)],
    service: Annotated[
        ServiceName | None,
        Query(description="The name of the service to filter users on"),
    ] = None,
    deletion_status: Annotated[
        DeletionStatus | None,
        Query(description="The deletion status to filter users on"),
    ] = None,
    offset: Annotated[
        int | None,
        Query(ge=0, description="The number of items to offset"),
    ] = 0,
    limit: Annotated[
        int | None,
        Query(le=1000, description="The maximum number of items to retrieve"),
    ] = 100,
) -> list[UserRead]:
    """Retrieve a list of users based on the query parameters."""
    statement = select(User)

    if service or deletion_status:
        statement = statement.join(UserServiceStatus)

    if service:
        statement = statement.where(UserServiceStatus.service_name == service)

    if deletion_status:
        statement = statement.where(UserServiceStatus.status == deletion_status)

    users = session.scalars(statement.offset(offset).limit(limit)).unique().all()

    response_users = [UserRead.model_validate(user) for user in users]
    logger.debug("Results = %s", response_users)
    return response_users


@router.get("/{user_id}")
async def read_user(
    session: Annotated[Session, Depends(get_session)],
    user_id: Annotated[UUID, Path(description="The id of the user to read")],
    service: Annotated[
        ServiceName | None,
        Query(description="The name of the service to filter users on"),
    ] = None,
) -> UserRead:
    """Retrieve the user from its id."""
    statement = select(User).where(User.id == user_id)

    if service:
        statement = statement.join(UserServiceStatus).where(
            UserServiceStatus.service_name == service
        )

    user = session.scalar(statement)

    if not user:
        message = "User not found"
        logger.debug("%s: %s", message, user_id)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    response_user = UserRead.model_validate(user)
    logger.debug("Result = %s", response_user)
    return response_user


@router.get("/{user_id}/status/{service_name}")
async def read_user_status(
    session: Annotated[Session, Depends(get_session)],
    user_id: Annotated[
        UUID, Path(description="The ID of the user to read status from")
    ],
    service_name: Annotated[
        ServiceName,
        Path(description="The name of the service making the request"),
    ],
) -> UserStatusRead:
    """Read the user deletion status for a specific service."""
    statement = select(UserServiceStatus).where(
        UserServiceStatus.user_id == user_id,
        UserServiceStatus.service_name == service_name,
    )

    service_status = session.scalar(statement)

    if not service_status:
        message = "User status not found"
        logger.debug("%s: %s %s", message, user_id, service_name)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    response = UserStatusRead(
        id=service_status.user_id,
        service_name=service_status.service_name,
        status=service_status.status,
    )
    logger.debug("Results = %s", response)

    return response


@router.patch("/{user_id}/status/{service_name}")
async def update_user_status(
    session: Annotated[Session, Depends(get_session)],
    user_id: Annotated[UUID, Path(title="The ID of the user to update status")],
    service_name: Annotated[
        ServiceName,
        Path(description="The name of the service to update status"),
    ],
    deletion_status: Annotated[
        DeletionStatus,
        Body(description="The new deletion status", embed=True),
    ],
) -> UserStatusUpdate:
    """Update the user deletion status for a specific service."""
    statement = (
        update(UserServiceStatus)
        .where(
            UserServiceStatus.user_id == user_id,
            UserServiceStatus.service_name == service_name,
        )
        .values(status=deletion_status)
        .returning(UserServiceStatus)
    )

    try:
        updated = session.execute(statement).scalars().one()
    except NoResultFound as exc:
        message = "User status not found"
        logger.debug("%s: %s", message, user_id)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=message
        ) from exc

    session.commit()

    response_user = UserStatusUpdate(
        id=updated.user_id,
        service_name=updated.service_name,
        status=updated.status,
    )
    logger.debug("Results = %s", response_user)

    return response_user



"""Mork edx MongoDB models."""

from mongoengine import (
    BooleanField,
    DateField,
    DictField,
    Document,
    IntField,
    ListField,
    ObjectIdField,
    StringField,
)
from mongoengine.queryset import queryset_manager


class ForumObject(Document):
    """Base model for a forum object document."""

    _id = ObjectIdField()
    votes = DictField()
    visible = BooleanField()
    abuse_flaggers = ListField()
    historical_abuse_flaggers = ListField()
    thread_type = StringField()
    context = StringField()
    comment_count = IntField()
    at_position_list = ListField()
    body = StringField()
    course_id = StringField()
    commentable_id = StringField()
    anonymous = BooleanField()
    anonymous_to_peers = BooleanField()
    closed = BooleanField()
    author_id = IntField()
    author_username = StringField()
    updated_at = DateField()
    created_at = DateField()
    last_activity_at = DateField()
    meta = {"abstract": True, "collection": "contents"}


class Comment(ForumObject):
    """Model for the `Comment` document type."""

    type = StringField(default="Comment", db_field="_type")

    @queryset_manager
    def objects(self, queryset):
        """Filter on the _type field."""
        return queryset.filter(type="Comment")


class CommentThread(ForumObject):
    """Model for the `CommentThread` document type."""

    type = StringField(default="CommentThread", db_field="_type")
    title = StringField()

    @queryset_manager
    def objects(self, queryset):
        """Filter on the _type field."""
        return queryset.filter(type="CommentThread")



"""Factory classes for util models."""

import factory

from mork.edx.mysql.models.util import UtilRatelimitconfiguration

from .base import BaseSQLAlchemyModelFactory


class EdxUtilRatelimitconfigurationFactory(BaseSQLAlchemyModelFactory):
    """Model for the `util_ratelimitconfiguration` table."""

    class Meta:
        """Factory configuration."""

        model = UtilRatelimitconfiguration

    id = factory.Sequence(lambda n: n + 1)
    change_date = factory.Faker("date_time")
    changed_by_id = factory.Sequence(lambda n: n + 1)
    enabled = factory.Faker("random_int", min=0, max=1)



"""Mork celery configuration file."""

import sentry_sdk
from celery import Celery, signals
from sentry_sdk.scrubber import DEFAULT_PII_DENYLIST, EventScrubber

from mork import __version__
from mork.conf import settings

from .probe import LivenessProbe

app = Celery(
    "mork",
    include=[
        "mork.celery.tasks.deletion",
        "mork.celery.tasks.edx",
        "mork.celery.tasks.emailing",
    ],
)
app.steps["worker"].add(LivenessProbe)


def before_send(event, hint):  # noqa: ARG001
    """Remove user email from the event sent to Sentry."""
    if not event.get("breadcrumbs"):
        return event

    for breadcrumb in event["breadcrumbs"].get("values", []):
        data = breadcrumb.get("data", {})
        url = data.get("url", "")

        # Remove user email from Brevo request URL
        if "/contacts/" in url:
            data["url"] = url.replace(url.split("/contacts/")[-1], "[Filtered]")
            breadcrumb["data"] = data

    return event


@signals.celeryd_init.connect
def init_sentry(**_kwargs):
    """Initialize Sentry SDK on Celery startup."""
    if settings.SENTRY_DSN is not None:
        pii_denylist = DEFAULT_PII_DENYLIST + ["email", "username"]
        sentry_sdk.init(
            dsn=settings.SENTRY_DSN,
            enable_tracing=True,
            traces_sample_rate=settings.SENTRY_CELERY_TRACES_SAMPLE_RATE,
            release=__version__,
            environment=settings.SENTRY_EXECUTION_ENVIRONMENT,
            max_breadcrumbs=50,
            send_default_pii=False,
            event_scrubber=EventScrubber(pii_denylist=pii_denylist, recursive=True),
            before_send=before_send,
        )
        sentry_sdk.set_tag("application", "celery")


# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("mork.conf:settings", namespace="CELERY")



"""Mork edx course models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class CourseActionStateCoursererunstate(Base):
    """Model for the `course_action_state_coursererunstate` table."""

    __tablename__ = "course_action_state_coursererunstate"
    __table_args__ = (
        ForeignKeyConstraint(
            ["created_user_id"],
            ["auth_user.id"],
            name="created_user_id_refs_id_1334640c1744bdeb",
        ),
        ForeignKeyConstraint(
            ["updated_user_id"],
            ["auth_user.id"],
            name="updated_user_id_refs_id_1334640c1744bdeb",
        ),
        Index(
            "course_action_state_coursererun_course_key_cf5da77ed3032d6_uniq",
            "course_key",
            "action",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    created_time: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    updated_time: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    created_user_id: Mapped[int] = mapped_column(INTEGER(11), index=True, nullable=True)
    updated_user_id: Mapped[int] = mapped_column(INTEGER(11), index=True, nullable=True)
    course_key: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    action: Mapped[str] = mapped_column(String(100), nullable=False, index=True)
    state: Mapped[str] = mapped_column(String(50), nullable=False)
    should_display: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    message: Mapped[str] = mapped_column(String(1000), nullable=False)
    source_course_key: Mapped[str] = mapped_column(
        String(255), nullable=False, index=True
    )
    display_name: Mapped[str] = mapped_column(String(255), nullable=False)

    created_user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser",
        foreign_keys=[created_user_id],
        back_populates="course_action_state_coursererunstate_created_user",
    )
    updated_user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser",
        foreign_keys=[updated_user_id],
        back_populates="course_action_state_coursererunstate_updated_user",
    )


class CourseCreatorsCoursecreator(Base):
    """Model for the `course_creators_coursecreator` table."""

    __tablename__ = "course_creators_coursecreator"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )
    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, unique=True)
    state_changed: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    state: Mapped[str] = mapped_column(String(24), nullable=False)
    note: Mapped[str] = mapped_column(String(512), nullable=False)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="course_creators_coursecreator"
    )


class CourseGroupsCourseusergroupUsers(Base):
    """Model for the `course_groups_courseusergroup_users` table."""

    __tablename__ = "course_groups_courseusergroup_users"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index("courseusergroup_id", "courseusergroup_id", "user_id", unique=True),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    courseusergroup_id: Mapped[int] = mapped_column(
        INTEGER(11), nullable=False, index=True
    )
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="course_groups_courseusergroup_users"
    )


class CourseGroupsCohortmembership(Base):
    """Model for the `course_groups_cohortmembership` table."""

    __tablename__ = "course_groups_cohortmembership"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index(
            "course_groups_cohortmembership_user_id_395bddd0389ed7da_uniq",
            "user_id",
            "course_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    course_user_group_id: Mapped[int] = mapped_column(
        INTEGER(11), nullable=False, index=True
    )
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="course_groups_cohortmembership"
    )



"""Module for CRUD functions."""

from datetime import datetime
from logging import getLogger
from typing import Optional

from sqlalchemy import delete, distinct, select, union_all
from sqlalchemy.orm import Session, load_only
from sqlalchemy.sql.functions import count

from mork.edx.mysql.models.auth import AuthtokenToken, AuthUser
from mork.edx.mysql.models.certificates import (
    CertificatesCertificatehtmlviewconfiguration,
)
from mork.edx.mysql.models.contentstore import ContentstoreVideouploadconfig
from mork.edx.mysql.models.course import (
    CourseActionStateCoursererunstate,
    CourseCreatorsCoursecreator,
)
from mork.edx.mysql.models.dark import DarkLangDarklangconfig
from mork.edx.mysql.models.student import StudentCourseenrollmentallowed
from mork.edx.mysql.models.util import UtilRatelimitconfiguration
from mork.edx.mysql.models.verify import VerifyStudentHistoricalverificationdeadline
from mork.edx.mysql.models.wiki import WikiArticle, WikiArticlerevision
from mork.exceptions import UserNotFound, UserProtected

logger = getLogger(__name__)


def get_inactive_users_count(
    session: Session,
    threshold_date: datetime,
) -> int:
    """Get inactive users count from edx database.

    SELECT count(auth_user.id) FROM auth_user
    """
    query = (
        select(count(distinct(AuthUser.id)))
        .prefix_with("SQL_NO_CACHE", dialect="mysql")
        .filter(AuthUser.last_login < threshold_date)
    )
    return session.execute(query).scalar()


def get_inactive_users(
    session: Session,
    threshold_date: datetime,
    offset: Optional[int] = 0,
    limit: Optional[int] = 0,
) -> list[AuthUser]:
    """Get users from edx database who have not logged in for a specified period.

    SELECT auth_user.id,
        auth_user.username,
        auth_user.email,
        auth_user.is_staff,
        auth_user.is_superuser,
        auth_user.last_login,
    FROM auth_user LIMIT :param_1 OFFSET :param_2
    """
    query = (
        select(AuthUser)
        .prefix_with("SQL_NO_CACHE", dialect="mysql")
        .options(
            load_only(
                AuthUser.id,
                AuthUser.username,
                AuthUser.email,
                AuthUser.is_staff,
                AuthUser.is_superuser,
                AuthUser.last_login,
            ),
        )
        .filter(AuthUser.last_login < threshold_date)
        .offset(offset)
        .limit(limit)
    )
    return session.scalars(query).unique().all()


def get_user(session: Session, email: str) -> AuthUser:
    """Get a user entry based on the provided email.

    Parameters:
    session (Session): SQLAlchemy session object.
    email (str): The email of the user to get.
    """
    query = select(AuthUser).where(AuthUser.email == email)
    return session.execute(query).scalar()


def _has_protected_children(session: Session, user_id) -> bool:
    """Check if user has an entry in a protected children table."""
    union_statement = union_all(
        select(1).where(AuthtokenToken.user_id == user_id),
        select(1).where(
            CertificatesCertificatehtmlviewconfiguration.changed_by_id == user_id
        ),
        select(1).where(ContentstoreVideouploadconfig.changed_by_id == user_id),
        select(1).where(CourseActionStateCoursererunstate.created_user_id == user_id),
        select(1).where(CourseActionStateCoursererunstate.updated_user_id == user_id),
        select(1).where(CourseCreatorsCoursecreator.user_id == user_id),
        select(1).where(DarkLangDarklangconfig.changed_by_id == user_id),
        select(1).where(UtilRatelimitconfiguration.changed_by_id == user_id),
        select(1).where(
            VerifyStudentHistoricalverificationdeadline.history_user_id == user_id
        ),
        select(1).where(WikiArticle.owner_id == user_id),
        select(1).where(WikiArticlerevision.user_id == user_id),
    )

    # Execute the union query and check if any results exist
    result = session.execute(union_statement).scalars().first()
    return bool(result)


def delete_user(session: Session, email: str) -> None:
    """Delete a user entry based on the provided email and username.

    Parameters:
    session (Session): SQLAlchemy session object.
    email (str): The email of the user to delete.
    """
    user_to_delete = session.scalar(select(AuthUser).where(AuthUser.email == email))
    if not user_to_delete:
        msg = "User does not exist"
        logger.warning(msg)
        raise UserNotFound(msg)

    if _has_protected_children(session, user_to_delete.id):
        msg = "User is linked to a protected table and cannot be deleted"
        logger.warning(msg)
        raise UserProtected(msg)

    # Delete entries in student_courseenrollmentallowed table containing user email
    session.execute(
        delete(StudentCourseenrollmentallowed).where(
            StudentCourseenrollmentallowed.email == email
        )
    )

    # Delete user from auth_user table and all its children
    session.delete(user_to_delete)

    logger.debug(f"Deleting user {email=}")



"""Mork database connection."""

import logging
from threading import Lock
from typing import Generator, Optional

from pydantic import PostgresDsn
from sqlalchemy import Engine as SAEngine
from sqlalchemy import NullPool, create_engine, text
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session as SASession

from mork.conf import settings

logger = logging.getLogger(__name__)


class MorkDB:
    """Class to connect to the Mork database.

    This class is used by Celery workers to start a new engine and session per process.
    """

    session = None

    def __init__(self):
        """Initialize SqlAlchemy engine and session."""
        # Disable pooling as SQLAlchemy connections cannot be shared accross processes,
        # and Celery forks processes by default
        self.engine = create_engine(
            settings.DB_URL, echo=settings.DB_DEBUG, poolclass=NullPool
        )
        self.session = SASession(self.engine)


class Singleton(type):
    """Thread-safe singleton pattern metaclass."""

    _instances: dict = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        """Store instances in a private class property."""
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Engine(metaclass=Singleton):
    """Database engine singleton used by the API to connect to the Mork database."""

    _engine: Optional[SAEngine] = None

    def get_engine(self, url: PostgresDsn, echo: bool = False) -> SAEngine:
        """Get created engine or create a new one."""
        if self._engine is None:
            logger.debug("Create a new engine")
            self._engine = create_engine(str(url), echo=echo)
        logger.debug("Getting database engine %s", self._engine)
        return self._engine


def get_engine() -> SAEngine:
    """Get database engine."""
    return Engine().get_engine(url=settings.DB_URL, echo=settings.DB_DEBUG)


def get_session() -> Generator[SASession, None, None]:
    """Get database session."""
    with SASession(bind=get_engine()) as session:
        logger.debug("Getting session %s", session)
        yield session
        logger.debug("Closing session %s", session)


def is_alive(session: SASession) -> bool:
    """Check if database connection is alive."""
    try:
        session.execute(text("SELECT 1 as is_alive"))
        return True
    except OperationalError as err:
        logger.debug("Exception: %s", err)
        return False



"""Factory classes for notify models."""

import factory

from mork.edx.mysql.models.notify import NotifySetting

from .base import BaseSQLAlchemyModelFactory


class EdxNotifySettingFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `notify_settings` table."""

    class Meta:
        """Factory configuration."""

        model = NotifySetting

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    interval = factory.Sequence(lambda n: n)



"""Mork models."""

from datetime import datetime

from sqlalchemy import DateTime, func
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    """Base class for all models in the database."""

    filtered_attrs = []

    def safe_dict(self):
        """Return a dictionary representation of the model."""
        return {
            c.name: getattr(self, c.name)
            for c in self.__table__.columns
            if c.name not in self.filtered_attrs
        }

    created_at: Mapped[datetime] = mapped_column(
        DateTime, server_default=func.now(), nullable=False
    )

    updated_at: Mapped[datetime] = mapped_column(
        DateTime, default=func.now(), onupdate=func.now(), nullable=False
    )



"""Tests for Mork Celery utils."""

import re

from pytest_httpx import HTTPXMock
from sqlalchemy import select

from mork.celery.utils import (
    get_service_status,
    get_user_from_mork,
    update_status_in_mork,
)
from mork.factories.users import UserFactory, UserServiceStatusFactory
from mork.models.users import DeletionStatus, ServiceName, User
from mork.schemas.users import UserRead


def test_get_user_from_mork(db_session, httpx_mock: HTTPXMock):
    """Test the behavior of getting a user from Mork API."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get user from db
    expected_user = UserRead.model_validate(db_session.scalar(select(User)))

    httpx_mock.add_response(
        url=re.compile(rf".*v1/users/{expected_user.id}"),
        method="GET",
        json=expected_user.model_dump(mode="json"),
        status_code=200,
    )

    user = get_user_from_mork(expected_user.id)

    assert user == expected_user


def test_get_user_from_mork_invalid(db_session, httpx_mock: HTTPXMock):
    """Test the behavior of getting a user from Mork with invalid user id or status."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get user from db
    expected_user = UserRead.model_validate(db_session.scalar(select(User)))

    httpx_mock.add_response(
        url=re.compile(rf".*v1/users/{expected_user.id}"),
        method="GET",
        status_code=404,
    )

    assert get_user_from_mork(expected_user.id) is None


def test_get_service_status(db_session, httpx_mock: HTTPXMock):
    """Test the behavior of getting the status from a user."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    expected_status = DeletionStatus.DELETED
    UserFactory.create(
        service_statuses={ServiceName.EDX: expected_status},
    )

    # Get user from db
    expected_user = UserRead.model_validate(db_session.scalar(select(User)))

    status = get_service_status(expected_user, ServiceName.EDX)

    assert status == expected_status


def test_get_service_status_not_found(db_session, httpx_mock: HTTPXMock):
    """Test the behavior of getting a inexistent status from a user."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get user from db
    expected_user = UserRead.model_validate(db_session.scalar(select(User)))

    # Remove statuses of the user
    expected_user.service_statuses = []

    status = get_service_status(expected_user, ServiceName.EDX)

    assert status is None


def test_update_status_in_mork(db_session, httpx_mock: HTTPXMock):
    """Test the behavior of updating the user status."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get id of one of the newly created user
    user_id = db_session.scalar(select(User.id))

    service_name = "ashley"

    httpx_mock.add_response(
        url=re.compile(rf".*v1/users/{user_id}/status/{service_name}"),
        method="PATCH",
        status_code=200,
    )

    success = update_status_in_mork(user_id, ServiceName.ASHLEY, DeletionStatus.DELETED)
    assert success


def test_update_user_status_invalid(db_session, httpx_mock: HTTPXMock):
    """Test the behavior of updating the user status with an invalid response."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get id of one of the newly created user
    user_id = db_session.scalar(select(User.id))

    service_name = "ashley"

    httpx_mock.add_response(
        url=re.compile(rf".*v1/users/{user_id}/status/{service_name}"),
        method="PATCH",
        status_code=404,
    )

    success = update_status_in_mork(user_id, ServiceName.ASHLEY, DeletionStatus.DELETED)
    assert not success



"""Mork Celery deletion tasks."""

from datetime import datetime
from logging import getLogger
from uuid import UUID

from celery import chain, group
from sqlalchemy import insert, select
from sqlalchemy.exc import DBAPIError, SQLAlchemyError

from mork.celery.celery_app import app
from mork.celery.tasks.brevo import delete_brevo_platform_user
from mork.celery.tasks.edx import delete_edx_platform_user
from mork.conf import settings
from mork.db import MorkDB
from mork.edx.mysql import crud
from mork.edx.mysql.database import OpenEdxMySQLDB
from mork.exceptions import UserDeleteError
from mork.models.tasks import EmailStatus
from mork.models.users import (
    DeletionReason,
    DeletionStatus,
    ServiceName,
    User,
    UserServiceStatus,
)

logger = getLogger(__name__)


@app.task
def delete_inactive_users(limit: int = 0, dry_run: bool = True):
    """Celery task to delete inactive users accounts."""
    edx_db = OpenEdxMySQLDB()

    threshold_date = datetime.now() - settings.DELETION_PERIOD
    total = crud.get_inactive_users_count(edx_db.session, threshold_date)

    if limit:
        total = min(total, limit)

    for batch_offset in range(0, total, settings.EDX_MYSQL_QUERY_BATCH_SIZE):
        batch_limit = min(settings.EDX_MYSQL_QUERY_BATCH_SIZE, total - batch_offset)
        inactive_users = crud.get_inactive_users(
            edx_db.session,
            threshold_date,
            offset=batch_offset,
            limit=batch_limit,
        )

        delete_group = group(
            [
                delete_user.s(
                    email=user.email, reason=DeletionReason.GDPR, dry_run=dry_run
                )
                for user in inactive_users
            ]
        )
        delete_group.delay()

    edx_db.session.close()


@app.task
def delete_user(
    email: str,
    reason: DeletionReason = DeletionReason.USER_REQUESTED,
    dry_run: bool = True,
):
    """Celery task that deletes a specific user."""
    if dry_run:
        logger.info("Dry run: User would have been deleted")
        return

    logger.debug(f"Starting deletion of user with {email=}")

    delete_chain = chain(
        remove_email_status.si(email),
        mark_user_for_deletion.si(email=email, reason=reason),
        group(delete_edx_platform_user.s(), delete_brevo_platform_user.s()),
    )

    delete_chain.delay()


@app.task
def mark_user_for_deletion(email: str, reason: DeletionReason) -> UUID:
    """Mark user for deletion across all services in Mork database."""
    logger.debug("Marking user for deletion")

    edx_mysql_db = OpenEdxMySQLDB()
    edx_mysql_db.session.expire_on_commit = False

    try:
        auth_user = crud.get_user(edx_mysql_db.session, email)
    except (SQLAlchemyError, DBAPIError) as exc:
        edx_mysql_db.session.rollback()
        msg = "Failed to read user from edX MySQL"
        logger.error(msg)
        raise UserDeleteError(msg) from exc
    finally:
        edx_mysql_db.session.close()

    if not auth_user:
        msg = "User not found in edx database"
        logger.error(msg)
        raise UserDeleteError(msg)

    mork_db = MorkDB()

    # Check if user already exist in Mork database
    user_exist = mork_db.session.scalar(select(User).where(User.email == email))
    if user_exist:
        logger.info("User is already marked for deletion")
        return user_exist.id

    # Add user to Mork database
    user_id = mork_db.session.scalar(
        insert(User)
        .returning(User.id)
        .values(
            username=auth_user.username,
            edx_user_id=auth_user.id,
            email=auth_user.email,
            reason=reason,
        )
    )

    # Mark for deletion across all services
    mork_db.session.execute(
        insert(UserServiceStatus),
        [
            {
                "user_id": user_id,
                "service_name": service,
                "status": DeletionStatus.TO_DELETE,
            }
            for service in ServiceName
        ],
    )

    try:
        mork_db.session.commit()
    except (SQLAlchemyError, DBAPIError) as exc:
        mork_db.session.rollback()
        logger.error(f"Failed to mark user to be deleted - {exc}")
        raise UserDeleteError("Failed to mark user to be deleted") from exc
    finally:
        mork_db.session.close()

    return user_id


@app.task
def remove_email_status(email: str):
    """Delete the email status in the Mork database."""
    logger.debug("Removing user email status")
    mork_db = MorkDB()
    user_to_delete = (
        mork_db.session.query(EmailStatus).filter(EmailStatus.email == email).first()
    )
    if not user_to_delete:
        logger.warning("Email status not found")
        return

    try:
        mork_db.session.delete(user_to_delete)
        mork_db.session.commit()
    except (SQLAlchemyError, DBAPIError):
        mork_db.session.rollback()
        logger.error("Failed to delete email status")
        return
    finally:
        mork_db.session.close()



"""Tests for Mork database module."""

import pytest
from sqlalchemy import text
from sqlalchemy.exc import OperationalError


def test_database_connection(db_session):
    """Test the PostgreSQL database connection."""
    try:
        db_session.execute(text("SELECT 42 as life"))
    except OperationalError:
        pytest.fail("Cannot connect to configured database")



"""Factory classes for student models."""

import factory

from mork.edx.mysql.models.student import (
    StudentAnonymoususerid,
    StudentCourseaccessrole,
    StudentCourseenrollment,
    StudentCourseenrollmentallowed,
    StudentCourseenrollmentattribute,
    StudentHistoricalcourseenrollment,
    StudentLanguageproficiency,
    StudentLoginfailure,
    StudentManualenrollmentaudit,
    StudentPendingemailchange,
    StudentUserstanding,
)

from .base import BaseSQLAlchemyModelFactory, faker


class EdxStudentAnonymoususeridFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_anonymoususerid` table."""

    class Meta:
        """Factory configuration."""

        model = StudentAnonymoususerid

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    anonymous_user_id = factory.Faker("hexify", text="^" * 32)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")


class EdxStudentCourseaccessroleFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_courseaccessrole` table."""

    class Meta:
        """Factory configuration."""

        model = StudentCourseaccessrole

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    org = factory.Faker("word")
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    role = factory.Faker("word")


class EdxStudentCourseenrollmentallowedFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_courseenrollmentallowed` table."""

    class Meta:
        """Factory configuration."""

        model = StudentCourseenrollmentallowed

    id = factory.Sequence(lambda n: n + 1)
    email = factory.Faker("email")
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    created = factory.Faker("date_time")
    auto_enroll = factory.Faker("random_int", min=0, max=1)


class EdxStudentCourseenrollmentattributeFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_courseenrollmentattribute` table."""

    class Meta:
        """Factory configuration."""

        model = StudentCourseenrollmentattribute

    id = factory.Sequence(lambda n: n + 1)
    enrollment_id = factory.Sequence(lambda n: n + 1)
    namespace = factory.Faker("pystr")
    name = factory.Faker("pystr")
    value = factory.Faker("pystr")


class EdxStudentHistoricalcourseenrollmentFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_historicalcourseenrollment` table."""

    class Meta:
        """Factory configuration."""

        model = StudentHistoricalcourseenrollment

    id = factory.Sequence(lambda n: n + 1)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    created = factory.Faker("date_time")
    is_active = factory.Faker("random_int", min=0, max=1)
    mode = factory.Faker("pystr")
    user_id = factory.Sequence(lambda n: n + 1)
    history_id = factory.Sequence(lambda n: n + 1)
    history_date = factory.Faker("date_time")
    history_user_id = factory.Sequence(lambda n: n + 1)
    history_type = factory.Faker("random_letter")


class EdxStudentLanguageproficiencyFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_languageproficiency` table."""

    class Meta:
        """Factory configuration."""

        model = StudentLanguageproficiency

    id = factory.Sequence(lambda n: n + 1)
    user_profile_id = factory.Sequence(lambda n: n + 1)
    code = factory.Faker("pystr", max_chars=16)


class EdxStudentLoginfailureFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_loginfailure` table."""

    class Meta:
        """Factory configuration."""

        model = StudentLoginfailure

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    failure_count = factory.Sequence(lambda n: n)
    lockout_until = factory.Faker("date_time")


class EdxStudentManualenrollmentauditFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_manualenrollmentaudit` table."""

    class Meta:
        """Factory configuration."""

        model = StudentManualenrollmentaudit

    id = factory.Sequence(lambda n: n + 1)
    enrollment_id = factory.Sequence(lambda n: n + 1)
    enrolled_by_id = factory.Sequence(lambda n: n + 1)
    enrolled_email = factory.Faker("email")
    time_stamp = factory.Faker("date_time")
    state_transition = factory.Faker("pystr")
    reason = factory.Faker("text")


class EdxStudentCourseenrollmentFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_courseenrollment` table."""

    class Meta:
        """Factory configuration."""

        model = StudentCourseenrollment

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    is_active = factory.Faker("random_int", min=0, max=1)
    mode = factory.Faker("word")
    created = factory.Faker("date_time")

    student_courseenrollmentattribute = factory.RelatedFactoryList(
        EdxStudentCourseenrollmentattributeFactory,
        "enrollment",
        size=3,
        enrollment_id=factory.SelfAttribute("..id"),
    )
    student_manualenrollmentaudit = factory.RelatedFactoryList(
        EdxStudentManualenrollmentauditFactory,
        "enrollment",
        size=3,
        enrollment_id=factory.SelfAttribute("..id"),
        enrolled_by_id=factory.SelfAttribute("..user_id"),
    )


class EdxStudentPendingemailchangeFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_pendingemailchange` table."""

    class Meta:
        """Factory configuration."""

        model = StudentPendingemailchange

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    new_email = factory.Faker("email")
    activation_key = factory.Faker("hexify", text="^" * 32)


class EdxStudentUserstandingFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `student_userstanding` table."""

    class Meta:
        """Factory configuration."""

        model = StudentUserstanding

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    account_status = factory.Faker("word")
    changed_by_id = factory.Sequence(lambda n: n + 1)
    standing_last_changed_at = factory.Faker("date_time")



"""API tasks router."""

import logging
from typing import Union

from celery.result import AsyncResult
from fastapi import APIRouter, Body, Depends, Response, status

from mork.auth import authenticate_api_key
from mork.schemas.tasks import (
    TASK_TYPE_TO_FUNC,
    DeleteInactiveUsers,
    DeleteUser,
    EmailInactiveUsers,
    EmailUser,
    TaskResponse,
    TaskStatus,
    TaskType,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/tasks", dependencies=[Depends(authenticate_api_key)])


@router.post("", status_code=status.HTTP_202_ACCEPTED)
@router.post("/", status_code=status.HTTP_202_ACCEPTED)
async def create_task(
    response: Response,
    task: Union[DeleteInactiveUsers, EmailInactiveUsers, DeleteUser, EmailUser] = Body(
        discriminator="type"
    ),
) -> TaskResponse:
    """Create a new task."""
    celery_task = TASK_TYPE_TO_FUNC[task.type]
    celery_params = task.model_dump(exclude="type", exclude_none=True)

    result = celery_task.delay(**celery_params)

    task_response = TaskResponse(id=result.task_id, status=TaskStatus.PENDING)
    response.headers["location"] = router.url_path_for(
        "get_task_status", **{"task_id": task_response.id}
    )
    return task_response


@router.options("")
@router.options("/")
async def get_available_tasks(response: Response) -> dict:
    """Get available tasks that can be created."""
    response.headers["allow"] = "POST"
    return {"task_types": list(TaskType)}


@router.get("/{task_id}/status")
async def get_task_status(task_id: str) -> TaskResponse:
    """Get the task status for `task_id`."""
    status = AsyncResult(task_id).state

    return TaskResponse(id=task_id, status=status)



"""Tests for Mork Celery Brevo tasks."""

import logging
import uuid
from unittest.mock import Mock
from uuid import uuid4

import httpx
import pytest
from sqlalchemy import select

from mork.celery.tasks.brevo import (
    delete_brevo_platform_user,
    delete_brevo_user,
)
from mork.conf import settings
from mork.exceptions import (
    UserDeleteError,
    UserNotFound,
    UserStatusError,
)
from mork.factories.users import UserFactory, UserServiceStatusFactory
from mork.models.users import DeletionStatus, ServiceName, User
from mork.schemas.users import UserRead


def test_delete_brevo_platform_user(db_session, monkeypatch):
    """Test to delete user from Brevo platform."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.brevo.get_user_from_mork", lambda x: user)

    mock_delete_brevo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.update_status_in_mork", mock_update_status_in_mork
    )

    delete_brevo_platform_user(user.id)

    mock_delete_brevo_user.assert_called_once_with(email=user.email)
    mock_update_status_in_mork.assert_called_once_with(
        user_id=user.id, service=ServiceName.BREVO, status=DeletionStatus.DELETED
    )


def test_delete_brevo_platform_user_empty_setting(db_session, monkeypatch):
    """Test to delete user from Brevo platform when the API URL is not set."""

    monkeypatch.setattr("mork.celery.tasks.brevo.settings.BREVO_API_URL", "")

    mock_delete_brevo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )

    delete_brevo_platform_user(uuid.uuid4())

    mock_delete_brevo_user.assert_not_called()


def test_delete_brevo_platform_user_invalid_user(monkeypatch):
    """Test to delete user from Brevo platform with an invalid user."""

    monkeypatch.setattr("mork.celery.tasks.brevo.get_user_from_mork", lambda x: None)

    mock_delete_brevo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.update_status_in_mork", mock_update_status_in_mork
    )

    nonexistent_id = uuid4().hex
    with pytest.raises(
        UserNotFound, match=f"User {nonexistent_id} could not be retrieved from Mork"
    ):
        delete_brevo_platform_user(nonexistent_id)

    mock_delete_brevo_user.assert_not_called()
    mock_update_status_in_mork.assert_not_called()


def test_delete_brevo_platform_user_deleted_status(db_session, monkeypatch, caplog):
    """Test to delete user from Brevo platform already deleted."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is already deleted on brevo
    UserFactory.create(
        service_statuses={ServiceName.BREVO: DeletionStatus.DELETED},
    )

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.brevo.get_user_from_mork", lambda x: user)

    mock_delete_brevo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.update_status_in_mork", mock_update_status_in_mork
    )

    # User is already deleted, silently exit the task
    with caplog.at_level(logging.WARNING):
        delete_brevo_platform_user(user.id)

    assert (
        "mork.celery.tasks.brevo",
        logging.WARNING,
        f"User {str(user.id)} has already been deleted.",
    ) in caplog.record_tuples

    mock_delete_brevo_user.assert_not_called()
    mock_update_status_in_mork.assert_not_called()


def test_delete_brevo_platform_user_invalid_status(db_session, monkeypatch, caplog):
    """Test to delete user from Brevo platform with an invalid status."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is being deleted on brevo
    UserFactory.create(
        service_statuses={ServiceName.BREVO: DeletionStatus.DELETING},
    )

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.brevo.get_user_from_mork", lambda x: user)

    mock_delete_brevo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.update_status_in_mork", mock_update_status_in_mork
    )

    with pytest.raises(
        UserStatusError,
        match=f"User {str(user.id)} is not to be deleted. Status: DeletionStatus.DELETING",  # noqa: E501
    ):
        delete_brevo_platform_user(user.id)

    mock_delete_brevo_user.assert_not_called()
    mock_update_status_in_mork.assert_not_called()


def test_delete_brevo_platform_user_failed_delete(db_session, monkeypatch):
    """Test to delete user from Brevo platform with a failed delete."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is already deleted on brevo
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.brevo.get_user_from_mork", lambda x: user)

    def mock_delete_brevo_user(*args, **kwars):
        raise UserDeleteError("An error occurred")

    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )

    with pytest.raises(UserDeleteError, match="An error occurred"):
        delete_brevo_platform_user(user.id)


def test_delete_brevo_platform_user_failed_status_update(db_session, monkeypatch):
    """Test to delete user from Brevo platform with a failed status update."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is already deleted on brevo
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.brevo.get_user_from_mork", lambda x: user)

    mock_delete_brevo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.delete_brevo_user", mock_delete_brevo_user
    )
    mock_update_status_in_mork = Mock(return_value=False)
    monkeypatch.setattr(
        "mork.celery.tasks.brevo.update_status_in_mork", mock_update_status_in_mork
    )

    with pytest.raises(
        UserStatusError,
        match=f"Failed to update deletion status to deleted for user {user.id}",
    ):
        delete_brevo_platform_user(user.id)


def test_delete_brevo_user(httpx_mock):
    """Test to delete user's data from Brevo."""

    email = "johndoe@example.com"

    httpx_mock.add_response(
        url=f"{settings.BREVO_API_URL}/contacts/{email}",
        method="DELETE",
        headers={"api-key": "not-a-real-api-key"},
        status_code=200,
    )

    delete_brevo_user(email)


def test_delete_brevo_user_request_error(monkeypatch):
    """Test to delete user's data from Brevo with a request error."""

    def mock_httpx_delete(*args, **kwars):
        raise httpx.RequestError("An error occurred")

    monkeypatch.setattr("mork.celery.tasks.brevo.httpx.delete", mock_httpx_delete)

    with pytest.raises(
        UserDeleteError, match="Network error while deleting user contact on Brevo"
    ):
        delete_brevo_user("johndoe@example.com")


def test_delete_brevo_user_not_found(monkeypatch, caplog):
    """Test to delete user's data from Brevo when no contact found."""

    def mock_httpx_delete(*args, **kwars):
        raise httpx.HTTPStatusError(
            "An error occured",
            request=None,
            response=httpx.Response(status_code=httpx.codes.NOT_FOUND),
        )

    monkeypatch.setattr("mork.celery.tasks.brevo.httpx.delete", mock_httpx_delete)

    # Make sure no error is raised if contact not found on Brevo
    with caplog.at_level(logging.INFO):
        delete_brevo_user("johndoe@example.com")

    assert (
        "mork.celery.tasks.brevo",
        logging.INFO,
        "User has no contact on Brevo",
    ) in caplog.record_tuples


def test_delete_brevo_user_status_error(monkeypatch):
    """Test to delete user's data from Brevo with API returning a 4** or 5**."""

    def mock_httpx_delete(*args, **kwars):
        raise httpx.HTTPStatusError(
            "An error occured",
            request=None,
            response=httpx.Response(status_code=httpx.codes.METHOD_NOT_ALLOWED),
        )

    monkeypatch.setattr("mork.celery.tasks.brevo.httpx.delete", mock_httpx_delete)

    with pytest.raises(UserDeleteError, match="Failed to delete user contact on Brevo"):
        delete_brevo_user("johndoe@example.com")



"""Mork edx courseware models."""

import datetime
from typing import List

from sqlalchemy import DateTime, Float, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class CoursewareOfflinecomputedgrade(Base):
    """Model for the `courseware_offlinecomputedgrade` table."""

    __tablename__ = "courseware_offlinecomputedgrade"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index(
            "courseware_offlinecomputedgrade_user_id_46133bbd0926078f_uniq",
            "user_id",
            "course_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, index=True)
    updated: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    gradeset: Mapped[str] = mapped_column(TEXT)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="courseware_offlinecomputedgrade"
    )


class CoursewareStudentmodule(Base):
    """Model for the `courseware_studentmodule` table."""

    __tablename__ = "courseware_studentmodule"
    __table_args__ = (
        ForeignKeyConstraint(
            ["student_id"],
            ["auth_user.id"],
            name="student_id_refs_id_51af713179ba2570",
        ),
        Index(
            "courseware_studentmodule_student_id_635d77aea1256de5_uniq",
            "student_id",
            "module_id",
            "course_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    module_type: Mapped[str] = mapped_column(String(32), nullable=False, index=True)
    module_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    student_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    state: Mapped[str] = mapped_column(TEXT)
    grade: Mapped[float] = mapped_column(Float(asdecimal=True), index=True)
    created: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    modified: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    max_grade: Mapped[float] = mapped_column(Float(asdecimal=True))
    done: Mapped[str] = mapped_column(String(8), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)

    student: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="courseware_studentmodule"
    )

    courseware_studentmodulehistory: Mapped[List["CoursewareStudentmodulehistory"]] = (
        relationship(
            "CoursewareStudentmodulehistory",
            back_populates="student_module",
            cascade="all, delete-orphan",
        )
    )


class CoursewareStudentmodulehistory(Base):
    """Model for the `courseware_studentmodulehistory` table."""

    __tablename__ = "courseware_studentmodulehistory"
    __table_args__ = (
        ForeignKeyConstraint(
            ["student_module_id"],
            ["courseware_studentmodule.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    student_module_id: Mapped["CoursewareStudentmodule"] = mapped_column(
        INTEGER(11), nullable=False, index=True
    )
    version: Mapped[str] = mapped_column(String(255), index=True)
    created: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    state: Mapped[str] = mapped_column(TEXT)
    grade: Mapped[float] = mapped_column(Float(asdecimal=True))
    max_grade: Mapped[float] = mapped_column(Float(asdecimal=True))

    student_module: Mapped["CoursewareStudentmodule"] = relationship(
        "CoursewareStudentmodule", back_populates="courseware_studentmodulehistory"
    )


class CoursewareXmodulestudentinfofield(Base):
    """Model for the `courseware_xmodulestudentinfofield` table."""

    __tablename__ = "courseware_xmodulestudentinfofield"
    __table_args__ = (
        ForeignKeyConstraint(
            ["student_id"],
            ["auth_user.id"],
        ),
        Index(
            "courseware_xmodulestudentinfof_student_id_33f2f772c49db067_uniq",
            "student_id",
            "field_name",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    field_name: Mapped[str] = mapped_column(String(64), nullable=False, index=True)
    value: Mapped[str] = mapped_column(TEXT, nullable=False)
    student_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    modified: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )

    student: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="courseware_xmodulestudentinfofield"
    )


class CoursewareXmodulestudentprefsfield(Base):
    """Model for the `courseware_xmodulestudentprefsfield` table."""

    __tablename__ = "courseware_xmodulestudentprefsfield"
    __table_args__ = (
        ForeignKeyConstraint(
            ["student_id"],
            ["auth_user.id"],
        ),
        Index(
            "courseware_xmodulestudentprefs_student_id_2a5d275498b7a407_uniq",
            "student_id",
            "module_type",
            "field_name",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    field_name: Mapped[str] = mapped_column(String(64), nullable=False, index=True)
    module_type: Mapped[str] = mapped_column(String(64), nullable=False, index=True)
    value: Mapped[str] = mapped_column(TEXT, nullable=False)
    student_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )
    modified: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False, index=True
    )

    student: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="courseware_xmodulestudentprefsfield"
    )



"""API health router."""

import logging
from enum import Enum
from typing import Annotated

from fastapi import APIRouter, Depends, Response, status
from pydantic import BaseModel
from sqlalchemy.orm import Session

from mork.db import get_session
from mork.db import is_alive as is_db_alive

logger = logging.getLogger(__name__)

router = APIRouter()


class DatabaseStatus(Enum):
    """Data backend statuses."""

    OK = "ok"
    AWAY = "away"
    ERROR = "error"


class Heartbeat(BaseModel):
    """Warren backends status."""

    database: DatabaseStatus

    @property
    def is_alive(self):
        """A helper that checks the overall status."""
        if self.database == DatabaseStatus.OK:
            return True
        return False


@router.get("/__lbheartbeat__")
async def lbheartbeat() -> None:
    """Load balancer heartbeat.

    Return a 200 when the server is running.
    """
    return


@router.get("/__heartbeat__", status_code=status.HTTP_200_OK)
async def heartbeat(
    session: Annotated[Session, Depends(get_session)], response: Response
) -> Heartbeat:
    """Application heartbeat.

    Return a 200 if all checks are successful.
    """
    statuses = Heartbeat(
        database=DatabaseStatus.OK if is_db_alive(session) else DatabaseStatus.ERROR,
    )
    if not statuses.is_alive:
        response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    return statuses



"""Mork Celery custom liveness and readiness probes."""

from pathlib import Path

from celery import bootsteps
from celery.signals import worker_ready, worker_shutdown

HEARTBEAT_FILE = Path("/tmp/worker_heartbeat")  # noqa: S108
READINESS_FILE = Path("/tmp/worker_ready")  # noqa: S108


class LivenessProbe(bootsteps.StartStopStep):
    """Celery worker component that implements a liveness probe mechanism."""

    requires = {"celery.worker.components:Timer"}

    def __init__(self, worker, **kwargs):  # noqa: ARG002
        """Initialize the liveness probe."""
        self.requests = []
        self.tref = None

    def start(self, worker):
        """Start the liveness probe with a periodic heartbeat."""
        # Touch the heartbeat file every second
        self.tref = worker.timer.call_repeatedly(
            1.0,
            self.update_heartbeat_file,
            (worker,),
            priority=10,
        )

    def stop(self, worker):  # noqa: ARG002
        """Stop the liveness probe by removing the heartbeat file."""
        HEARTBEAT_FILE.unlink(missing_ok=True)

    def update_heartbeat_file(self, worker):  # noqa: ARG002
        """Update the heartbeat file by touching it."""
        HEARTBEAT_FILE.touch()


@worker_ready.connect
def worker_ready(**_):
    """Signal handler that creates a readiness file when the worker is ready."""
    READINESS_FILE.touch()


@worker_shutdown.connect
def worker_shutdown(**_):
    """Signal handler that removes the readiness file when the worker shuts down."""
    READINESS_FILE.unlink(missing_ok=True)



"""Factory classes for proctoru models."""

import factory

from mork.edx.mysql.models.proctoru import ProctoruProctoruexam, ProctoruProctoruuser

from .base import BaseSQLAlchemyModelFactory


class EdxProctoruProctoruexamFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `proctoru_proctoruexam` table."""

    class Meta:
        """Factory configuration."""

        model = ProctoruProctoruexam

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    start_date = factory.Faker("date_time")
    actual_start_time = factory.Faker("date_time")
    is_completed = factory.Faker("random_int", min=0, max=1)
    is_started = factory.Faker("random_int", min=0, max=1)
    is_canceled = factory.Faker("random_int", min=0, max=1)
    block_id = factory.Faker("hexify", text="^" * 32)
    end_time = factory.Faker("date_time")
    reservation_id = factory.Faker("hexify", text="^" * 40)
    reservation_no = str(factory.Sequence(lambda n: n))
    url = factory.Faker("url")


class EdxProctoruProctoruuserFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `proctoru_proctoruuser` table."""

    class Meta:
        """Factory configuration."""

        model = ProctoruProctoruuser

    id = factory.Sequence(lambda n: n + 1)
    student_id = factory.Sequence(lambda n: n + 1)
    phone_number = factory.Faker("phone_number")
    time_zone = factory.Faker("timezone")
    address = factory.Faker("street_address")
    city = factory.Faker("city")
    country = factory.Faker("country_code")
    date_created = factory.Faker("date_time")
    time_zone_display_name = factory.Faker("pystr", max_chars=100)
    state = factory.Faker("word")



"""Factory classes for instructor models."""

import factory

from mork.edx.mysql.models.instructor import InstructorTaskInstructortask

from .base import BaseSQLAlchemyModelFactory, faker


class EdxInstructorTaskInstructortaskFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `instructor_task_instructortask` table."""

    class Meta:
        """Factory configuration."""

        model = InstructorTaskInstructortask

    id = factory.Sequence(lambda n: n + 1)
    task_type = factory.Faker("word")
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    task_key = factory.Faker("hexify", text="^" * 20)
    task_input = factory.Faker("pystr", max_chars=255)
    task_id = factory.Faker("uuid4")
    task_state = factory.Faker("word")
    task_output = factory.Faker("json")
    requester_id = factory.Sequence(lambda n: n + 1)
    created = factory.Faker("date_time")
    updated = factory.Faker("date_time")
    subtasks = factory.Faker("json")



"""Factory classes for verify models."""

import factory

from mork.edx.mysql.models.wiki import WikiArticle, WikiArticlerevision

from .base import BaseSQLAlchemyModelFactory


class WikiArticleFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `wiki_article` table."""

    class Meta:
        """Factory configuration."""

        model = WikiArticle

    id = factory.Sequence(lambda n: n + 1)
    current_revision_id = factory.Sequence(lambda n: n + 1)
    created = factory.Faker("date_time")
    modified = factory.Faker("date_time")
    owner_id = factory.Sequence(lambda n: n + 1)
    group_id = factory.Sequence(lambda n: n + 1)
    group_read = factory.Faker("random_int", min=0, max=1)
    group_write = factory.Faker("random_int", min=0, max=1)
    other_read = factory.Faker("random_int", min=0, max=1)
    other_write = factory.Faker("random_int", min=0, max=1)


class WikiArticlerevisionFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `wiki_articlerevision` table."""

    class Meta:
        """Factory configuration."""

        model = WikiArticlerevision

    id = factory.Sequence(lambda n: n + 1)
    revision_number = factory.Sequence(lambda n: n + 1)
    user_message = factory.Faker("text")
    automatic_log = factory.Faker("text")
    ip_address = ""
    user_id = factory.Sequence(lambda n: n + 1)
    modified = factory.Faker("date_time")
    created = factory.Faker("date_time")
    previous_revision_id = factory.Sequence(lambda n: n + 1)
    deleted = factory.Faker("random_int", min=0, max=1)
    locked = factory.Faker("random_int", min=0, max=1)
    article_id = factory.Sequence(lambda n: n + 1)
    content = factory.Faker("text", max_nb_chars=2000)
    title = factory.Faker("sentence")



"""Add timestamp fields to email tables

Revision ID: 976e462b45b6
Revises: 608d075c6e99
Create Date: 2024-11-05 16:27:23.065141

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "976e462b45b6"
down_revision: Union[str, None] = "608d075c6e99"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "email_status",
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
    )
    op.add_column(
        "email_status", sa.Column("updated_at", sa.DateTime(), nullable=False)
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("email_status", "updated_at")
    op.drop_column("email_status", "created_at")
    # ### end Alembic commands ###



"""Mork edx wiki models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class WikiArticle(Base):
    """Model for the `wiki_article` table."""

    __tablename__ = "wiki_article"
    __table_args__ = (
        ForeignKeyConstraint(
            ["owner_id"],
            ["auth_user.id"],
        ),
    )
    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    current_revision_id: Mapped[int] = mapped_column(INTEGER(11), unique=True)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    modified: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    owner_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    group_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    group_read: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    group_write: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    other_read: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    other_write: Mapped[int] = mapped_column(INTEGER(1), nullable=False)

    owner: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="wiki_article"
    )


class WikiArticlerevision(Base):
    """Model for the `wiki_articlerevision` table."""

    __tablename__ = "wiki_articlerevision"
    __table_args__ = (
        Index(
            "wiki_articlerevision_article_id_4b4e7910c8e7b2d0_uniq",
            "article_id",
            "revision_number",
            unique=True,
        ),
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    revision_number: Mapped[int] = mapped_column(INTEGER(11), nullable=False)
    user_message: Mapped[str] = mapped_column(TEXT, nullable=False)
    automatic_log: Mapped[str] = mapped_column(TEXT, nullable=False)
    ip_address: Mapped[str] = mapped_column(String(15))
    user_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    modified: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    previous_revision_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    deleted: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    locked: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    article_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    content: Mapped[str] = mapped_column(TEXT, nullable=False)
    title: Mapped[str] = mapped_column(String(512), nullable=False)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="wiki_articlerevision"
    )
